# -*- coding: utf-8 -*-

class TreeProperty(object):

    def __init__(self, name):
        self.name = name
        self._name = '_' + name

    def __set__(self, instance, value):
        instance.__dict__[self._name] = value

    def __get__(self, instance, cls):
        if not hasattr(instance, self._name):
            instance.__dict__[self._name] = getattr(instance._parent, self.name)
        return instance.__dict__[self._name]





# -*- coding: utf-8 -*-

import six

from .base import SheetBase, BookBase
from .writermixin import SheetMixin, BookMixin, Box
from .utils import tag_test, parse_cell_tag
from .xlnode import Tree, Row, Cell, EmptyCell, Node, create_cell
from .jinja import JinjaEnv
from .nodemap import NodeMap
from .sheetresource import SheetResourceMap
from .richtexthandler import rich_handler
from .merger import Merger
from .config import config
from .celltag import CellTag

class SheetWriter(SheetBase, SheetMixin):

    def __init__(self, bookwriter, sheet_resource, sheet_name):
        self.rdbook = bookwriter.rdbook
        self.wtbook = bookwriter.wtbook
        self.style_list = bookwriter.style_list
        self.merger = sheet_resource.merger
        self.rdsheet = sheet_resource.rdsheet
        self.create_worksheet(self.rdsheet, sheet_name)
        self.wtrows = set()
        self.wtcols = set()
        self.box = Box(-1, -1)

class BookWriter(BookBase, BookMixin):
    sheet_writer_cls = SheetWriter

    def __init__(self, fname, debug=False):
        config.debug = debug
        self.load(fname)

    def load(self, fname):
        self.workbook = self.load_rdbook(fname)
        self.font_map = {}
        self.node_map = NodeMap()
        self.jinja_env = JinjaEnv(self.node_map)
        self.merger_cls = Merger
        self.sheet_writer_map = {}
        self.sheet_resource_map = SheetResourceMap(self, self.jinja_env)
        for index,rdsheet in enumerate(self.rdbook.sheets()):
            self.sheet_resource_map.add(rdsheet, rdsheet.name, index)

    def build(self, sheet, index, merger):
        tree = Tree(index, self.node_map)
        for rowx in range(sheet.nrows):
            row_node = Row(rowx)
            tree.add_child(row_node)
            for colx in range(sheet.ncols):
                try:
                    sheet_cell = sheet.cell(rowx, colx)
                except:
                    cell_node = EmptyCell(rowx, colx)
                    tree.add_child(cell_node)
                    continue
                cell_tag_map = None
                note = sheet.cell_note_map.get((rowx, colx))
                if note:
                    comment = note.text
                    if tag_test(comment):
                        _, cell_tag_map = parse_cell_tag(comment)
                value = sheet_cell.value
                cty = sheet_cell.ctype
                rich_text = self.get_rich_text(sheet, rowx, colx)
                if isinstance(value, six.text_type):
                    if not tag_test(value):
                        if rich_text:
                            cell_node = Cell(sheet_cell, rowx, colx, rich_text, cty)
                        else:
                            cell_node = Cell(sheet_cell, rowx, colx, value, cty)
                    else:
                        font = self.get_font(sheet, rowx, colx)
                        cell_node = create_cell(sheet_cell, rowx, colx, value, rich_text, cty, font, rich_handler)
                else:
                    cell_node = Cell(sheet_cell, rowx, colx, value, cty)
                if cell_tag_map:
                    cell_tag = CellTag(cell_tag_map)
                    cell_node.extend_cell_tag(cell_tag)
                    if colx==0:
                        row_node.cell_tag = cell_tag
                tree.add_child(cell_node)
        tree.add_child(Node())#
        return tree

    def render_sheet(self, payload, left_top=None):
        if not hasattr(self, 'wtbook') or self.wtbook is None:
            self.create_workbook()
        return BookMixin.render_sheet(self, payload, left_top)

    def save(self, fname):
        if self.wtbook is not None:
            stream = open(fname, 'wb')
            self.wtbook.save(stream)
            stream.close()
            del self.wtbook
        self.sheet_writer_map.clear()



# -*- coding: utf-8 -*-

import os
import six
from jinja2 import nodes
from jinja2.ext import Extension
from jinja2.runtime import Undefined
from inspect import isfunction

class NodeExtension(Extension):
    tags = set(['row', 'cell', 'node', 'extra'])

    def parse(self, parser):
        lineno = next(parser.stream).lineno
        args = [parser.parse_expression()]
        body = []
        return nodes.CallBlock(self.call_method('_node', args),
                               [], [], body).set_lineno(lineno)

    def _node(self, key, caller):
        node = self.environment.node_map.get_node(key)
        return str(key)

class SegmentExtension(Extension):
    tags = set(['seg'])

    def parse(self, parser):
        lineno = next(parser.stream).lineno
        args = [parser.parse_expression()]
        body = parser.parse_statements(['name:endseg'], drop_needle=True)
        return nodes.CallBlock(self.call_method('_seg', args),
                               [], [], body).set_lineno(lineno)

    def _seg(self, key, caller):
        segment = self.environment.node_map.get_node(key)
        rv = caller()
        rv = segment.process_rv(rv)
        return rv

class XvExtension(Extension):
    tags = set(['xv'])

    def parse(self, parser):
        lineno = next(parser.stream).lineno
        args = [parser.parse_expression()]
        if parser.stream.skip_if('comma'):
            args.append(parser.parse_expression())
        else:
            args.append(nodes.Const(0))
        body = []
        return nodes.CallBlock(self.call_method('_xv', args),
                               [], [], body).set_lineno(lineno)

    def _xv(self, xv, key, caller):
        if key==0:
            return six.text_type(xv)
        xvcell = self.environment.node_map.get_node(key)
        if xv is None or type(xv) is Undefined:
            xv = ''
        xvcell.rv = xv
        return six.text_type(xv)

class OpExtension(Extension):
    tags = set(['op'])

    def parse(self, parser):
        lineno = next(parser.stream).lineno
        args = [parser.parse_expression()]
        func_args = []
        while parser.stream.skip_if('comma'):
            func_args.append(parser.parse_expression())
        args.append(nodes.List(func_args))
        body = []
        return nodes.CallBlock(self.call_method('_op', args),
                               [], [], body).set_lineno(lineno)

    def _op(self, func, func_args, caller):
        if(isfunction(func)):
            node = self.environment.node_map.current_node
            node.add_op((func, func_args))
        return six.text_type(func)

class NoopExtension(Extension):
    tags = set(['op'])

    def parse(self, parser):
        lineno = next(parser.stream).lineno
        args = [parser.parse_expression()]
        func_args = []
        while parser.stream.skip_if('comma'):
            func_args.append(parser.parse_expression())
        args.append(nodes.List(func_args))
        body = []
        return nodes.CallBlock(self.call_method('_op', args),
                               [], [], body).set_lineno(lineno)

    def _op(self, func, func_args, caller):
        return six.text_type(func)

class ImageExtension(Extension):
    tags = set(['img'])

    def parse(self, parser):
        lineno = next(parser.stream).lineno
        args = [parser.parse_expression()]
        if parser.stream.skip_if('comma'):
            args.append(parser.parse_expression())
        else:
            args.append(nodes.Const(0))
        body = []
        return nodes.CallBlock(self.call_method('_image', args),
                               [], [], body).set_lineno(lineno)

    def _image(self, image_ref, image_key, caller):
         return ''

try:
    pil = True
    from PIL.ImageFile import ImageFile
except:
    pil = False

class ImageRef():

    def __init__(self, image, image_index):
        self.image = image
        self.image_index = image_index
        self.rdrowx = -1
        self.rdcolx = -1
        self.wtrowx = -1
        self.wtcolx = -1
        if not isinstance(image, ImageFile):
            fname = six.text_type(image)
            if not os.path.exists(fname):
                self.image = None

    @property
    def image_key(self):
        return (self.rdrowx,self.rdcolx,self.image_index)

    @property
    def wt_top_left(self):
        return (self.wtrowx,self.wtcolx)

class ImagexExtension(Extension):
    tags = set(['img'])

    def parse(self, parser):
        lineno = next(parser.stream).lineno
        args = [parser.parse_expression()]
        if parser.stream.skip_if('comma'):
            args.append(parser.parse_expression())
        else:
            args.append(nodes.Const(0))
        body = []
        return nodes.CallBlock(self.call_method('_image', args),
                               [], [], body).set_lineno(lineno)

    def _image(self, image, image_index, caller):
        if not pil:
            return ''
        image_ref = ImageRef(image, image_index)
        if image_ref.image:
            node = self.environment.node_map.current_node
            node.set_image_ref(image_ref)
        return 'image'



# -*- coding: utf-8 -*-

import copy
from jinja2 import nodes
from jinja2.ext import Extension
from openpyxl.cell.rich_text import TextBlock

def yes(font):
    wfont = copy.copy(font)
    wfont.name = 'Wingdings 2'
    return 'R', wfont

def yesx(font):
    wfont = copy.copy(font)
    wfont.rFont = 'Wingdings 2'
    return TextBlock(wfont, 'R')

def no():
    return u'□'

def yn(value, font, xlsx):
    if value:
        if xlsx:
            return yesx(font)
        else:
            return yes(font)
    else:
        return no()

class YnExtension(Extension):
    tags = set(['yn'])
    xlsx = False

    def parse(self, parser):
        lineno = next(parser.stream).lineno
        args = [parser.parse_expression()]
        if parser.stream.skip_if('comma'):
            args.append(parser.parse_expression())
        else:
            args.append(nodes.Const(None))
        body = []
        return nodes.CallBlock(self.call_method('_yn', args),
                               [], [], body).set_lineno(lineno)

    def _yn(self, arg0, arg1, caller):
        segment = self.environment.node_map.current_node
        if arg1 is not None:
            arg0 = not arg0
        rv = yn(arg0, segment.font, self.xlsx)
        rv = segment.process_rich_rv(rv)
        return rv

class YnxExtension(YnExtension):
    xlsx = True



class Box(object):

    def __init__(self, top, left):
        self.reset_pos(top, left)

    def next_row(self):
        self.bottom += 1
        self.right = self.left

    def next_cell(self):
        self.right += 1

    def reset_pos(self, top, left):
        self.top = top
        self.bottom = top
        self.left = left
        self.right = left

class SheetMixin(object):

    def copy_dimensions(self, rdrowx, rdcolx, wtrowx, wtcolx):
        self.copy_row_dimension(rdrowx, wtrowx)
        self.copy_col_dimension(rdcolx, wtcolx)

    def reset_pos(self, top_left):
        if isinstance(top_left, (tuple, list)):
            self.box = Box(top_left[0], top_left[1])
        elif isinstance(top_left, dict):
            self.box = Box(top_left['top'], top_left['left'])
        else:
            self.box = Box(top_left.top, top_left.left)

    def set_sheet_resource(self, sheet_resource):
        self.merger = sheet_resource.merger
        self.rdsheet = sheet_resource.rdsheet

    def write_row(self, row_node):
        self.box.next_row()

    def write_cell(self, cell_node, rv, cty):
        self.box.next_cell()
        self.merger.merge_cell(cell_node.rowx, cell_node.colx, self.box.bottom, self.box.right)
        self.copy_dimensions(cell_node.rowx, cell_node.colx, self.box.bottom, self.box.right)
        if cell_node.sheet_cell:
            cell_context = self.get_cell_context(cell_node, rv, cty)
            cell_context.finish()
            target_cell = cell_context.target_cell
            if target_cell and hasattr(cell_node, 'ops') and cell_node.ops:
                for (func, func_args) in cell_node.ops:
                    func(*func_args, cell_context)
                cell_node.ops.clear()

    def set_image_ref(self, image_ref):
        image_ref.wtrowx = self.box.bottom
        image_ref.wtcolx = self.box.right + 1
        self.merger.set_image_ref(image_ref)


class BookMixin(object):

    def load(self, fname):
        pass

    def build(self, sheet, index):
        pass

    def add_filter(self, key, value):
        self.jinja_env.filters[key] = value

    def add_global(self, key, value):
        self.jinja_env.globals[key] = value

    def set_jinja_filters(self, **kwargs):
        self.jinja_env.filters.update(kwargs)

    def set_jinja_globals(self, **kwargs):
        self.jinja_env.globals.update(kwargs)

    def get_sheet_writer(self, sheet_resource, sheet_name):
        sheet_writer = self.sheet_writer_map.get(sheet_name)
        if not sheet_writer:
            sheet_writer = self.sheet_writer_cls(self, sheet_resource, sheet_name)
            self.sheet_writer_map[sheet_name] = sheet_writer
        else:
            sheet_writer.set_sheet_resource(sheet_resource)
        return sheet_writer

    def get_sheet_name(self, payload):
        sheet_name = payload.get('sheet_name')
        if sheet_name:
            return sheet_name
        for i in range(9999):
            sheet_name = "sheet%d" % i
            if not self.sheet_writer_map.get(sheet_name):
                return sheet_name
        return "XLSheet"

    def get_sheet_resource(self, payload):
        return self.sheet_resource_map.get_sheet_resource(payload)

    def render_sheet(self, payload, top_left=None):
        sheet_name = self.get_sheet_name(payload)
        sheet_resource = self.get_sheet_resource(payload)
        sheet_writer = self.get_sheet_writer(sheet_resource, sheet_name)
        if top_left:
            sheet_writer.reset_pos(top_left)
        sheet_resource.render_sheet(sheet_writer, payload)
        return sheet_writer.box

    def render_sheets(self, payloads):
        for payload in payloads:
            self.render_sheet(payload)

    def render_book(self, payloads):
        return self.render_sheets(payloads)

    def save(self, fname):
        pass


# -*- coding: utf-8 -*-

class MergeMixin():

    def set_range(self, rdrowx=-1, rdcolx=-1, wtrowx=-1, wtcolx=-1):
        self.start_rdrowx = rdrowx
        self.start_rdcolx = rdcolx
        self.start_wtrowx = wtrowx
        self.start_wtcolx = wtcolx
        self.end_wtrowx = wtrowx
        self.end_wtcolx = wtcolx

    def is_in_range(self, rdrowx, rdcolx):
        return self._first_row <= rdrowx <= self._last_row and \
               self._first_col <= rdcolx <= self._last_col

    def to_be_merged(self, rdrowx, rdcolx):
        if rdrowx > self.start_rdrowx:
            return True
        else:
            return rdrowx == self.start_rdrowx and rdcolx > self.start_rdcolx

    def merge_cell(self, rdrowx, rdcolx, wtrowx, wtcolx):
        if not self.is_in_range(rdrowx, rdcolx):
            return False
        if self.start_rdrowx == -1:
            self.set_range(rdrowx, rdcolx, wtrowx, wtcolx)
        elif self.to_be_merged(rdrowx, rdcolx):
            self.end_wtrowx = max(self.end_wtrowx, wtrowx)
            self.end_wtcolx = max(self.end_wtcolx, wtcolx)
        else:
            self.new_range()
            self.set_range(rdrowx, rdcolx, wtrowx, wtcolx)
        return True

    def new_range(self):
        pass

    def collect_range(self):
        self.new_range()
        self.set_range()

class CellMerge(MergeMixin):

    def __init__(self, cell_range, merger):
        self.merger = merger
        self.set_range()
        rlo, rhi, clo, chi = cell_range
        self._first_row = rlo
        self._last_row = rhi - 1
        self._first_col = clo
        self._last_col = chi - 1

    def new_range(self):
        if self.start_wtrowx==self.end_wtrowx and self.start_wtcolx==self.end_wtcolx:
            return
        range = (self.start_wtrowx, self.end_wtrowx, self.start_wtcolx, self.end_wtcolx)
        self.merger.add_new_range(range)

class Merger:

    def __init__(self, sheet):
        self.range_list = []
        self._merge_list = []
        self.get_merge_list(sheet)

    def get_merge_list(self, sheet):
        for range in sheet.merged_cells:
            _merge = CellMerge(range, self)
            self._merge_list.append(_merge)

    def add_new_range(self, range):
        self.range_list.append(range)

    def merge_cell(self, rdrowx, rdcolx, wtrowx, wtcolx):
        for _merge in self._merge_list:
            is_in_range = _merge.merge_cell(rdrowx, rdcolx, wtrowx, wtcolx)
            if is_in_range:
                break

    def collect_range(self, wtsheet):
        for _merge in self._merge_list:
            _merge.collect_range()
        for range in self.range_list:
            wtsheet.merged_ranges.append(range)
        self.range_list.clear()



from copy import deepcopy
from openpyxl.drawing.image import Image

# to avoid file closed error
class Cache():

    def __init__(self):
        self.map = {}

    def get_data(self, key):
        return self.map.get(key)

    def set_data(self, key, data):
        self.map[key] = data

    def clear(self):
        self.map.clear()


img_cache = Cache()
data_cache = Cache()

class Img(Image):

    def __init__(self, image):
        self.copy_ref(image)

    def copy_ref(self, image):
        self.anchor = deepcopy(image.anchor)
        self.ref = image.ref
        self.width = image.width
        self.height = image.height
        self.format = image.format

    def set_ref(self, ref):
        super(Img, self).__init__(ref)

    @property
    def path(self):
        key = self.key
        path = img_cache.get_data(key)
        if path:
            return path
        path = super(Img, self).path
        img_cache.set_data(key, path)
        return path

    def _data(self):
        key = self.key
        data = data_cache.get_data(key)
        if data:
            return data
        data = super(Img, self)._data()
        data_cache.set_data(key, data)
        return data

    @property
    def key(self):
        return id(self.ref)




